#include <Ultrasonic.h>   // https://github.com/ErickSimoes/Ultrasonic
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <ArduinoOTA.h>
#include <ArduinoJson.h>
#include <arduino_secrets.h>

// Device configuration
const String deviceName = "Water Softener";

const int minDistance = 10; // in cm
const int maxDistance = 21; // in cm

// WiFi configuration
const char* ssid = SECRET_SSID;
const char* password = SECRET_PASS;

// MQTT configuration
const char* mqttServer = SECRET_MQTT_BROKER;      // MQTT broker IP
const char* mqttClient = "water_softener";        // MQTT client name
const char* mqttUser = SECRET_MQTT_USERNAME;      // MQTT username
const char* mqttPassword = SECRET_MQTT_PASSWORD;  // MQTT password
const int mqttPort = 1883;                        // MQTT port (default: 1883)

Ultrasonic ultrasonic(13);
WiFiClient espClient;
PubSubClient client(espClient);

float prevMedian = 0;
float percentage;
unsigned long lastMQTTconnect = 0;
unsigned long previousMillis = 0; // Stores the last time measurements were taken
const long interval = 1000;       // Interval at which to take measurements (milliseconds)
int count = 0;                    // Counter for the number of measurements taken
float measurements[5];            // Array to store measurements

void autoDiscovery() {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[512];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["mdl"] = "Combilux I";
  deviceData["ids"] = mqttClient;
  deviceData["mf"] = "SmartHomeGeeks.io for Boshuis";
  deviceData["sw"] = "2024.06";

  DynamicJsonDocument root(512);

  // percentage sensor
  root["name"] = "Regeneration Salt Percentage";
  root["obj_id"] = String(mqttClient) + "_percentage";
  root["stat_t"] = String(mqttClient) + "/percentage";
  root["unit_of_meas"] = "%";
  root["icon"] = "mdi:shaker-outline";
  root["uniq_id"] = "sensor." + String(mqttClient) + "_percentage";
  root["avty_t"] = String(mqttClient) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqttClient) + "/percentage/config").c_str(), buffer, true);
  root.clear();

  // distance sensor
  root["name"] = "Regeneration Salt Distance";
  root["obj_id"] = String(mqttClient) + "_distance";
  root["stat_t"] = String(mqttClient) + "/distance";
  root["unit_of_meas"] = "cm";
  root["state_class"] = "measurement";
  root["icon"] = "mdi:waves-arrow-up";
  root["uniq_id"] = "sensor." + String(mqttClient) + "_distance";
  root["avty_t"] = String(mqttClient) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqttClient) + "/distance/config").c_str(), buffer, true);
  root.clear();

  // RSSI
  root["name"] = "RSSI";
  root["obj_id"] = String(mqttClient) + "_rssi";
  root["stat_t"] = String(mqttClient) + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + String(mqttClient) + "_rssi";
  root["avty_t"] = String(mqttClient) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqttClient) + "/rssi/config").c_str(), buffer, true);
  root.clear();

  // IP
  root["name"] = "IP";
  root["obj_id"] = String(mqttClient) + "_ip";
  root["stat_t"] = String(mqttClient) + "/ip";
  root["ic"] = "mdi:ip-network";
  root["uniq_id"] = "sensor." + String(mqttClient) + "_ip";
  root["avty_t"] = String(mqttClient) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqttClient) + "/ip/config").c_str(), buffer, true);
  root.clear();
}

void setup_wifi() {
  delay(1000);

  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.print(ssid);

  WiFi.mode(WIFI_STA);
  WiFi.hostname(mqttClient);
  WiFi.begin(ssid, password);

  unsigned long startAttemptTime = millis();
  while (WiFi.status() != WL_CONNECTED && millis() - startAttemptTime < 10000) {
    delay(500);
    Serial.print(".");
  }
  Serial.println();

  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("WiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }
}

void setup() {
  Serial.begin(115200);

  // Setup WiFi  
  setup_wifi();

  // Configure OTA
  ArduinoOTA.setHostname(mqttClient);
  ArduinoOTA.setPassword(SECRET_OTA_PASSWORD);
  ArduinoOTA.begin();

  client.setServer(mqttServer, mqttPort);
  connectmqtt();
}

void connectmqtt() {
  if (WiFi.status() != WL_CONNECTED) setup_wifi();

  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqttClient, mqttUser, mqttPassword, (String(mqttClient) + "/availability").c_str(), 1, true, "offline")) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);
    autoDiscovery();

    client.publish((String(mqttClient) + "/availability").c_str(), "online", true);
    client.publish((String(mqttClient) + "/ip").c_str(),  WiFi.localIP().toString().c_str(), true);

  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }
}

void loop() {
  if (!client.connected()) connectmqtt();
  
  unsigned long currentMillis = millis();

  // Check if it's time to take a measurement
  if (currentMillis - previousMillis >= interval) {
    previousMillis = currentMillis;  // Save the last time a measurement was taken

    if (count < 5) {
      measurements[count] = ultrasonic.read();  // Read distance and store it
      count++;
    } else {
      // Once 5 measurements are taken, compute the median
      float median = computeMedian(measurements, 5);
      if (abs(median - prevMedian) >= 1) {  // Check if median has changed significantly
        prevMedian = median;  // Update previous median

        Serial.print(median);
        Serial.println(" cm");

        // Calculate percentage
        float percentage = round(100 - (median - minDistance) / (maxDistance - minDistance) * 100);
        percentage = constrain(percentage, 0, 100);  // Ensure percentage is within 0 to 100

        // Publish the values
        client.publish((String(mqttClient) + "/distance").c_str(), String(median, 0).c_str(), true);
        client.publish((String(mqttClient) + "/percentage").c_str(), String(percentage, 0).c_str(), true);
        client.publish((String(mqttClient) + "/rssi").c_str(), String(WiFi.RSSI()).c_str(), true);
      }
      count = 0;  // Reset counter to start taking measurements again
    }
  }

  ArduinoOTA.handle();

  client.loop();
}

// Function to compute the median of an array of float values
float computeMedian(float arr[], int n) {
  simpleSort(arr, n); // Use a simple sort function
  if (n % 2 != 0)  // Check for odd number of elements
      return arr[n / 2];  // Return middle element
  return (arr[(n - 1) / 2] + arr[n / 2]) / 2.0;  // Return average of two middle elements
}

// Simple sorting function to sort array of floats
void simpleSort(float array[], int n) {
  for (int i = 0; i < n-1; i++) {
    for (int j = 0; j < n-i-1; j++) {
      if (array[j] > array[j+1]) {
        float temp = array[j];
        array[j] = array[j+1];
        array[j+1] = temp;
      }
    }
  }
}